@@ -13,13 +13,29 @@ typedef struct {
 
 
 typedef struct {
-    uint32_t     next_elt;
+    /* next_elt + property descriptor : 32 bits */
+
+    uint32_t     next_elt:26;
+
+    uint32_t     type:3;
+    uint32_t     writable:1;
+    uint32_t     enumerable:1;
+    uint32_t     configurable:1;
+
     uint32_t     key_hash;
-    void         *value;
+
+    void         *value[16 / sizeof(void *)];
 } njs_flathsh_elt_t;
 
 
-typedef struct njs_flathsh_descr_s  njs_flathsh_descr_t;
+typedef struct {
+    uint32_t     hash_mask;
+    uint32_t     elts_size;          /* allocated properties */
+    uint32_t     elts_count;         /* include deleted properties */
+    uint32_t     elts_deleted_count;
+} njs_flathsh_descr_t;
+
+
 typedef struct njs_flathsh_query_s  njs_flathsh_query_t;
 
 typedef njs_int_t (*njs_flathsh_test_t)(njs_flathsh_query_t *fhq, void *data);
@@ -64,6 +80,10 @@ struct njs_flathsh_query_s {
     ((fhl)->slot == (fhr)->slot)
 
 
+#define njs_hash_elts(h)                                                       \
+    ((njs_flathsh_elt_t *) ((char *) (h) + sizeof(njs_flathsh_descr_t)))
+
+
 /*
  * njs_flathsh_find() finds a hash element.  If the element has been
  * found then it is stored in the fhq->value and njs_flathsh_find()
@@ -73,16 +93,20 @@ struct njs_flathsh_query_s {
  */
 NJS_EXPORT njs_int_t njs_flathsh_find(const njs_flathsh_t *fh,
     njs_flathsh_query_t *fhq);
+/*
+ * The same as njs_flathsh_find(), but for hashes with unique keys.
+ */
+NJS_EXPORT njs_int_t njs_flathsh_unique_find(const njs_flathsh_t *fh,
+    njs_flathsh_query_t *fhq);
 
 /*
- * njs_flathsh_insert() adds a hash element.  If the element is already
- * present in flathsh and the fhq->replace flag is zero, then fhq->value
- * is updated with the old element and NJS_DECLINED is returned.
- * If the element is already present in flathsh and the fhq->replace flag
- * is non-zero, then the old element is replaced with the new element.
- * fhq->value is updated with the old element, and NJS_OK is returned.
+ * njs_flathsh_insert() adds a hash element.  If the element is already present
+ * in flathsh and the fhq->replace flag is zero, then NJS_DECLINED is returned.
+ * If the element is already present in flathsh and the fhq->replace flag is
+ * non-zero, then the old element is replaced with the new element and NJS_OK is
+ * returned.
  * If the element is not present in flathsh, then it is inserted and
- * NJS_OK is returned.  The fhq->value is not changed.
+ * NJS_OK is returned.
  * On memory allocation failure NJS_ERROR is returned.
  *
  * The required njs_flathsh_query_t fields: key_hash, key, proto, replace,
@@ -91,6 +115,11 @@ NJS_EXPORT njs_int_t njs_flathsh_find(const njs_flathsh_t *fh,
  */
 NJS_EXPORT njs_int_t njs_flathsh_insert(njs_flathsh_t *fh,
     njs_flathsh_query_t *fhq);
+/*
+ * The same as njs_flathsh_insert(), but for hashes with unique keys.
+ */
+NJS_EXPORT njs_int_t njs_flathsh_unique_insert(njs_flathsh_t *fh,
+    njs_flathsh_query_t *fhq);
 
 /*
  * njs_flathsh_delete() deletes a hash element.  If the element has been
@@ -102,11 +131,14 @@ NJS_EXPORT njs_int_t njs_flathsh_insert(njs_flathsh_t *fh,
  */
 NJS_EXPORT njs_int_t njs_flathsh_delete(njs_flathsh_t *fh,
     njs_flathsh_query_t *fhq);
+/*
+ * The same as njs_flathsh_delete(), but for hashes with unique keys.
+ */
+NJS_EXPORT njs_int_t njs_flathsh_unique_delete(njs_flathsh_t *fh,
+    njs_flathsh_query_t *fhq);
 
 
 typedef struct {
-    const njs_flathsh_proto_t  *proto;
-    uint32_t                   key_hash;
     uint32_t                   cp;
 } njs_flathsh_each_t;
 
@@ -114,11 +146,10 @@ typedef struct {
 #define njs_flathsh_each_init(lhe, _proto)                                     \
     do {                                                                       \
         njs_memzero(lhe, sizeof(njs_flathsh_each_t));                          \
-        (lhe)->proto = _proto;                                                 \
     } while (0)
 
 
-NJS_EXPORT void *njs_flathsh_each(const njs_flathsh_t *fh,
+NJS_EXPORT njs_flathsh_elt_t *njs_flathsh_each(const njs_flathsh_t *fh,
     njs_flathsh_each_t *fhe);
 
 /*
@@ -151,7 +182,19 @@ typedef struct njs_flathsh_proto_s  njs_lvlhsh_proto_t;
 #define njs_lvlhsh_insert(lh, lhq) njs_flathsh_insert(lh, lhq)
 #define njs_lvlhsh_delete(lh, lhq) njs_flathsh_delete(lh, lhq)
 #define njs_lvlhsh_each_init(lhe, _proto)  njs_flathsh_each_init(lhe, _proto)
-#define njs_lvlhsh_each(lh, lhe) njs_flathsh_each(lh, lhe)
+
+njs_inline njs_flathsh_elt_t *
+njs_lvlhsh_each(const njs_flathsh_t *lh, njs_flathsh_each_t *lhe)
+{
+    njs_flathsh_elt_t  *e;
+
+    e = njs_flathsh_each(lh, lhe);
+    if (e == NULL) {
+        return NULL;
+    }
+
+    return e;
+}
 
 
 #endif /* _NJS_FLATHSH_H_INCLUDED_ */
