@@ -19,6 +19,7 @@ typedef enum {
     NJS_OBJECT_PROP_CONFIGURABLE = 16,
     NJS_OBJECT_PROP_WRITABLE = 32,
     NJS_OBJECT_PROP_UNSET = 64,
+    NJS_OBJECT_PROP_IS_STRING = 128,
 #define NJS_OBJECT_PROP_VALUE_ECW (NJS_OBJECT_PROP_VALUE                     \
                                    | NJS_OBJECT_PROP_ENUMERABLE              \
                                    | NJS_OBJECT_PROP_CONFIGURABLE            \
@@ -39,8 +40,8 @@ typedef enum {
 
 
 struct njs_object_init_s {
-    const njs_object_prop_t     *properties;
-    njs_uint_t                  items;
+    const njs_object_prop_init_t  *properties;
+    njs_uint_t                    items;
 };
 
 
@@ -49,6 +50,7 @@ typedef struct njs_traverse_s  njs_traverse_t;
 struct njs_traverse_s {
     struct njs_traverse_s      *parent;
     njs_object_prop_t          *prop;
+    uint32_t                   atom_id;
 
     njs_value_t                value;
     njs_array_t                *keys;
@@ -72,21 +74,24 @@ njs_array_t *njs_object_own_enumerate(njs_vm_t *vm, const njs_object_t *object,
     uint32_t flags);
 njs_int_t njs_object_traverse(njs_vm_t *vm, njs_object_t *object, void *ctx,
     njs_object_traverse_cb_t cb);
-njs_int_t njs_object_hash_create(njs_vm_t *vm, njs_lvlhsh_t *hash,
-    const njs_object_prop_t *prop, njs_uint_t n);
+njs_int_t njs_object_make_shared(njs_vm_t *vm, njs_object_t *object);
+njs_int_t njs_object_hash_create(njs_vm_t *vm, njs_flathsh_t *hash,
+    const njs_object_prop_init_t *prop, njs_uint_t n);
 njs_int_t njs_primitive_prototype_get_proto(njs_vm_t *vm,
-    njs_object_prop_t *prop, njs_value_t *value, njs_value_t *setval,
-    njs_value_t *retval);
+    njs_object_prop_t *prop, uint32_t unused, njs_value_t *value,
+    njs_value_t *setval, njs_value_t *retval);
 njs_int_t njs_object_prototype_create(njs_vm_t *vm, njs_object_prop_t *prop,
-    njs_value_t *value, njs_value_t *setval, njs_value_t *retval);
-njs_value_t *njs_property_prototype_create(njs_vm_t *vm, njs_lvlhsh_t *hash,
+    uint32_t unused, njs_value_t *value, njs_value_t *setval,
+    njs_value_t *retval);
+njs_value_t *njs_property_prototype_create(njs_vm_t *vm, njs_flathsh_t *hash,
     njs_object_t *prototype);
 njs_int_t njs_object_prototype_proto(njs_vm_t *vm, njs_object_prop_t *prop,
-    njs_value_t *value, njs_value_t *setval, njs_value_t *retval);
-njs_int_t njs_object_prototype_create_constructor(njs_vm_t *vm,
-    njs_object_prop_t *prop, njs_value_t *value, njs_value_t *setval,
+    uint32_t unused, njs_value_t *value, njs_value_t *setval,
     njs_value_t *retval);
-njs_value_t *njs_property_constructor_set(njs_vm_t *vm, njs_lvlhsh_t *hash,
+njs_int_t njs_object_prototype_create_constructor(njs_vm_t *vm,
+    njs_object_prop_t *prop, uint32_t unused, njs_value_t *value,
+    njs_value_t *setval, njs_value_t *retval);
+njs_value_t *njs_property_constructor_set(njs_vm_t *vm, njs_flathsh_t *hash,
     njs_value_t *constructor);
 njs_int_t njs_object_to_string(njs_vm_t *vm, njs_value_t *value,
     njs_value_t *retval);
@@ -96,30 +101,32 @@ njs_int_t njs_object_length(njs_vm_t *vm, njs_value_t *value, int64_t *dst);
 
 njs_int_t njs_prop_private_copy(njs_vm_t *vm, njs_property_query_t *pq,
     njs_object_t *proto);
-njs_object_prop_t *njs_object_prop_alloc(njs_vm_t *vm, const njs_value_t *name,
-    const njs_value_t *value, uint8_t attributes);
+void njs_object_prop_init(njs_object_prop_t *prop, njs_object_prop_type_t type,
+    uint8_t attributes);
 njs_int_t njs_object_property(njs_vm_t *vm, njs_object_t *object,
-    njs_lvlhsh_query_t *lhq, njs_value_t *retval);
+    njs_flathsh_query_t *lhq, njs_value_t *retval);
 njs_object_prop_t *njs_object_property_add(njs_vm_t *vm, njs_value_t *object,
-    njs_value_t *key, njs_bool_t replace);
+    uint32_t atom_id, njs_bool_t replace);
 njs_int_t njs_object_prop_define(njs_vm_t *vm, njs_value_t *object,
-    njs_value_t *name, njs_value_t *value, unsigned flags, uint32_t hash);
+    uint32_t atom_id, njs_value_t *value, uint32_t flags);
 njs_int_t njs_object_prop_descriptor(njs_vm_t *vm, njs_value_t *dest,
     njs_value_t *value, njs_value_t *setval);
 njs_int_t njs_object_get_prototype_of(njs_vm_t *vm, njs_value_t *args,
     njs_uint_t nargs, njs_index_t unused, njs_value_t *retval);
 const char *njs_prop_type_string(njs_object_prop_type_t type);
-njs_int_t njs_object_prop_init(njs_vm_t *vm, const njs_object_init_t* init,
-    const njs_object_prop_t *base, njs_value_t *value, njs_value_t *retval);
+njs_int_t njs_object_props_init(njs_vm_t *vm, const njs_object_init_t* init,
+    njs_object_prop_t *base, uint32_t atom_id, njs_value_t *value,
+    njs_value_t *retval);
 
 
 njs_inline njs_bool_t
 njs_is_data_descriptor(njs_object_prop_t *prop)
 {
-    return prop->writable != NJS_ATTRIBUTE_UNSET
-           || (prop->type != NJS_ACCESSOR && njs_is_valid(njs_prop_value(prop)))
-           || prop->type == NJS_PROPERTY_HANDLER;
-
+    return (prop->type == NJS_PROPERTY && njs_is_valid(njs_prop_value(prop)))
+            || prop->type == NJS_PROPERTY_HANDLER
+            || prop->type == NJS_PROPERTY_REF
+            || prop->type == NJS_PROPERTY_PLACE_REF
+            || prop->type == NJS_PROPERTY_TYPED_ARRAY_REF;
 }
 
 
@@ -130,84 +137,38 @@ njs_is_accessor_descriptor(njs_object_prop_t *prop)
 }
 
 
-njs_inline njs_bool_t
-njs_is_generic_descriptor(njs_object_prop_t *prop)
-{
-    return !njs_is_data_descriptor(prop) && !njs_is_accessor_descriptor(prop);
-}
-
-
-njs_inline void
-njs_object_property_key_set(njs_lvlhsh_query_t *lhq, const njs_value_t *key,
-    uint32_t hash)
-{
-    if (njs_is_symbol(key)) {
-
-        lhq->key.length = 0;
-        lhq->key.start = NULL;
-        lhq->key_hash = njs_symbol_key(key);
-
-    } else {
-
-        /* string. */
-
-        njs_string_get(key, &lhq->key);
-
-        if (hash == 0) {
-            lhq->key_hash = njs_djb_hash(lhq->key.start, lhq->key.length);
-
-        } else {
-            lhq->key_hash = hash;
-        }
-    }
-}
-
-
-njs_inline void
-njs_object_property_init(njs_lvlhsh_query_t *lhq, const njs_value_t *key,
-    uint32_t hash)
-{
-    lhq->proto = &njs_object_hash_proto;
-
-    njs_object_property_key_set(lhq, key, hash);
-}
-
-
 njs_inline njs_int_t
 njs_primitive_value_to_key(njs_vm_t *vm, njs_value_t *dst,
     const njs_value_t *src)
 {
-    const njs_value_t  *value;
-
     switch (src->type) {
-
     case NJS_NULL:
-        value = &njs_string_null;
-        break;
+        njs_atom_to_value(vm, dst, NJS_ATOM_STRING_null);
+        return NJS_OK;
 
     case NJS_UNDEFINED:
-        value = &njs_string_undefined;
-        break;
+        njs_atom_to_value(vm, dst, NJS_ATOM_STRING_undefined);
+        return NJS_OK;
 
     case NJS_BOOLEAN:
-        value = njs_is_true(src) ? &njs_string_true : &njs_string_false;
-        break;
+        if (njs_is_true(src)) {
+            njs_atom_to_value(vm, dst, NJS_ATOM_STRING_true);
+
+        } else {
+            njs_atom_to_value(vm, dst, NJS_ATOM_STRING_false);
+        }
+
+        return NJS_OK;
 
     case NJS_NUMBER:
-        return njs_number_to_string(vm, dst, src);
-
     case NJS_SYMBOL:
     case NJS_STRING:
-        value = src;
-        break;
+        *dst = *src;
+        return NJS_OK;
 
     default:
         return NJS_ERROR;
     }
-
-    *dst = *value;
-
-    return NJS_OK;
 }
 
 
@@ -260,21 +221,62 @@ njs_key_string_get(njs_vm_t *vm, njs_value_t *key, njs_str_t *str)
         if (njs_slow_path(ret != NJS_OK)) {
             return ret;
         }
+
+    } else if (njs_slow_path(!njs_is_string(key))) {
+        ret = njs_value_to_string(vm, key, key);
+        if (njs_slow_path(ret != NJS_OK)) {
+            return ret;
+        }
     }
 
-    njs_string_get(key, str);
+    njs_string_get(vm, key, str);
 
     return NJS_OK;
 }
 
 
+njs_inline void
+njs_atom_string_get(njs_vm_t *vm, uint32_t atom_id, njs_str_t *str)
+{
+    njs_value_t  value;
+
+    if (njs_atom_to_value(vm, &value, atom_id) != NJS_OK) {
+        str->start = (u_char *) "unknown";
+        str->length = njs_length("unknown");
+    }
+
+    njs_key_string_get(vm, &value, str);
+}
+
+
+njs_inline njs_int_t
+njs_object_prop_define_val(njs_vm_t *vm, njs_value_t *object, njs_value_t *name,
+    njs_value_t *value, unsigned flags)
+{
+    njs_int_t  ret;
+
+    if (njs_value_atom(name) == NJS_ATOM_STRING_unknown) {
+        ret = njs_atom_atomize_key(vm, name);
+        if (ret != NJS_OK) {
+            return ret;
+        }
+    }
+
+    if (njs_is_string(name)) {
+       flags |= NJS_OBJECT_PROP_IS_STRING;
+    }
+
+    return njs_object_prop_define(vm, object, name->atom_id, value, flags);
+}
+
+
 njs_inline njs_int_t
 njs_value_create_data_prop(njs_vm_t *vm, njs_value_t *value,
     njs_value_t *name, njs_value_t *setval, uint32_t hash)
 {
-    return njs_object_prop_define(vm, value, name, setval,
-                                  NJS_OBJECT_PROP_CREATE
-                                  | NJS_OBJECT_PROP_VALUE_ECW, hash);
+    return njs_object_prop_define_val(vm, value, name, setval,
+                                      NJS_OBJECT_PROP_CREATE
+                                      | NJS_OBJECT_PROP_VALUE_ECW);
 }
 
 
@@ -295,12 +297,9 @@ njs_object_length_set(njs_vm_t *vm, njs_value_t *value, int64_t length)
 {
     njs_value_t  index;
 
-    static const njs_value_t  string_length = njs_string("length");
-
     njs_value_number_set(&index, length);
 
-    return njs_value_property_set(vm, value, njs_value_arg(&string_length),
-                                  &index);
+    return njs_value_property_set(vm, value, NJS_ATOM_STRING_length, &index);
 }
 
 
@@ -309,10 +308,7 @@ njs_object_string_tag(njs_vm_t *vm, njs_value_t *value, njs_value_t *tag)
 {
     njs_int_t  ret;
 
-    static const njs_value_t  to_string_tag =
-                                njs_wellknown_symbol(NJS_SYMBOL_TO_STRING_TAG);
-
-    ret = njs_value_property(vm, value, njs_value_arg(&to_string_tag), tag);
+    ret = njs_value_property(vm, value, NJS_ATOM_SYMBOL_toStringTag, tag);
     if (njs_slow_path(ret != NJS_OK)) {
         return ret;
     }
